RHO_KGM3 = 1.2  # Air density (kg m-3)
CP_JKG = 1013  # Specific heat of air at constant pressure (J/kg/K)
GAMMA_HPA = 0.67  # Psychrometric constant (hpa/K)
PT_ALPHA = 1.26  # Priestley-Taylor coefficient
SB_SIGMA = 5.67e-8  # Stefann Boltzmann constant
DEFAULT_G_METHOD = "santanello"
